"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
"""

import base64
import httplib2
import mimetypes
import os

from email.encoders import encode_base64
from email.mime.application import MIMEApplication
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

from platypush.plugins import action
from platypush.plugins.google import GooglePlugin


class GoogleMailPlugin(GooglePlugin):
    """
    GMail plugin. It allows you to programmatically compose and (TODO) get emails
    """

    scopes = ['https://www.googleapis.com/auth/gmail.modify']

    def __init__(self, *args, **kwargs):
        super().__init__(scopes=self.scopes, *args, **kwargs)


    @action
    def compose(self, sender, to, subject, body, files=None):
        """
        Compose a message.

        :param sender: Sender email/name
        :type sender: str

        :param to: Recipient email or comma-separated list of recipient emails
        :type to: str

        :param subject: Email subject
        :type subject: str

        :param body: Email body
        :type body: str

        :param files: Optional list of files to attach
        :type files: list
        """

        message = MIMEMultipart() if files else MIMEText(body)
        message['to'] = to
        message['from'] = sender
        message['subject'] = subject

        if files:
            for file in files:
                msg = MIMEText(body)
                message.attach(msg)
                content_type, encoding = mimetypes.guess_type(file)

                if content_type is None or encoding is not None:
                    content_type = 'application/octet-stream'

                main_type, sub_type = content_type.split('/', 1)
                with open(file, 'rb') as fp: content = fp.read()

                if main_type == 'text':
                    msg = mimetypes.MIMEText(content, _subtype=sub_type)
                elif main_type == 'image':
                    msg = MIMEImage(content, _subtype=sub_type)
                elif main_type == 'audio':
                    msg = MIMEAudio(content, _subtype=sub_type)
                elif main_type == 'application':
                    msg = MIMEApplication(content, _subtype=sub_type,
                                          _encoder=encode_base64)
                else:
                    msg = MIMEBase(main_type, sub_type)
                    msg.set_payload(content)

                filename = os.path.basename(file)
                msg.add_header('Content-Disposition', 'attachment', filename=filename)
                message.attach(msg)

        service = self.get_service('gmail', 'v1')
        body = { 'raw': base64.urlsafe_b64encode(message.as_bytes()).decode() }
        message = (service.users().messages().send(
            userId='me', body=body).execute())

        return message


    @action
    def get_labels(self):
        """
        Returns the available labels on the GMail account
        """
        service = self.get_service('gmail', 'v1')
        results = service.users().labels().list(userId='me').execute()
        labels = results.get('labels', [])
        return labels


# vim:sw=4:ts=4:et:



from typing import Optional, Dict, Any

from platypush.message.event import Event


class ZwaveEvent(Event):
    def __init__(self, device: Optional[str] = None, *args, **kwargs):
        super().__init__(*args, device=device, **kwargs)


class ZwaveNetworkReadyEvent(ZwaveEvent):
    """
    Triggered when the network started on a Z-Wave adapter becomes ready.
    """
    def __init__(self,
                 ozw_library_version: str,
                 python_library_version: str,
                 zwave_library: str,
                 node_id: int,
                 node_version: str,
                 home_id: int,
                 nodes_count: int,
                 device: Optional[str] = None,
                 *args, **kwargs):
        super().__init__(*args,
                         device=device,
                         ozw_library_version=ozw_library_version,
                         python_library_version=python_library_version,
                         zwave_library=zwave_library,
                         home_id=home_id,
                         node_id=node_id,
                         node_version=node_version,
                         nodes_count=nodes_count,
                         **kwargs)


class ZwaveNetworkStoppedEvent(ZwaveEvent):
    """
    Triggered when a Z-Wave network is stopped.
    """
    pass


class ZwaveNetworkErrorEvent(ZwaveEvent):
    """
    Triggered when an error occurs on the Z-Wave network.
    """
    pass


class ZwaveNetworkResetEvent(ZwaveEvent):
    """
    Triggered when a Z-Wave network is reset.
    """
    pass


class ZwaveNodeEvent(ZwaveEvent):
    """
    Generic Z-Wave node event class.
    """
    def __init__(self, node: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, node=node, **kwargs)


class ZwaveNodeAddedEvent(ZwaveNodeEvent):
    """
    Triggered when a node is added to the network.
    """
    pass


class ZwaveNodeRemovedEvent(ZwaveNodeEvent):
    """
    Triggered when a node is removed from the network.
    """
    pass


class ZwaveNodeRenamedEvent(ZwaveNodeEvent):
    """
    Triggered when a node is renamed.
    """
    pass


class ZwaveNodeReadyEvent(ZwaveNodeEvent):
    """
    Triggered when a node is ready.
    """
    pass


class ZwaveNodeGroupEvent(ZwaveNodeEvent):
    """
    Triggered when a node is associated/de-associated to a group.
    """
    def __init__(self, group_index: Optional[int] = None, *args, **kwargs):
        super().__init__(*args, group_index=group_index, **kwargs)


class ZwaveNodeSceneEvent(ZwaveNodeEvent):
    """
    Triggered when a scene is activated on a node.
    """
    def __init__(self, scene_id: int, *args, **kwargs):
        super().__init__(*args, scene_id=scene_id, **kwargs)


class ZwaveNodePollingEnabledEvent(ZwaveNodeEvent):
    """
    Triggered when the polling of a node is successfully turned on.
    """
    pass


class ZwaveNodePollingDisabledEvent(ZwaveNodeEvent):
    """
    Triggered when the polling of a node is successfully turned off.
    """
    pass


class ZwaveButtonCreatedEvent(ZwaveNodeEvent):
    """
    Triggered when a button is added to the network.
    """
    pass


class ZwaveButtonRemovedEvent(ZwaveNodeEvent):
    """
    Triggered when a button is removed from the network.
    """
    pass


class ZwaveButtonOnEvent(ZwaveNodeEvent):
    """
    Triggered when a button is pressed.
    """
    pass


class ZwaveButtonOffEvent(ZwaveNodeEvent):
    """
    Triggered when a button is released.
    """
    pass


class ZwaveValueEvent(ZwaveEvent):
    """
    Abstract class for Z-Wave value events.
    """
    def __init__(self, node: Dict[str, Any], value: Dict[str, Any], *args, **kwargs):
        super().__init__(*args, node=node, value=value, **kwargs)


class ZwaveValueAddedEvent(ZwaveValueEvent):
    """
    Triggered when a value is added to a node on the network.
    """
    pass


class ZwaveValueChangedEvent(ZwaveValueEvent):
    """
    Triggered when a value of a node on the network changes.
    """
    pass


class ZwaveValueRefreshedEvent(ZwaveValueEvent):
    """
    Triggered when a value of a node on the network is refreshed.
    """
    pass


class ZwaveValueRemovedEvent(ZwaveValueEvent):
    """
    Triggered when a value of a node on the network is removed.
    """
    pass


class ZwaveNodeQueryCompletedEvent(ZwaveEvent):
    """
    Triggered when all the nodes on the network have been queried.
    """
    pass


class ZwaveCommandEvent(ZwaveEvent):
    """
    Triggered when a command is received on the network.
    """
    def __init__(self, state: str, state_description: str, error: Optional[str] = None,
                 error_description: Optional[str] = None, node: Optional[Dict[str, Any]] = None, *args, **kwargs):
        super().__init__(*args, state=state, state_description=state_description,
                         error=error, error_description=error_description, node=node, **kwargs)


class ZwaveCommandWaitingEvent(ZwaveCommandEvent):
    """
    Triggered when a command is waiting for a message to proceed.
    """
    pass


# vim:sw=4:ts=4:et:



from platypush.backend.stt import SttBackend


class SttPicovoiceHotwordBackend(SttBackend):
    """
    Backend for the PicoVoice hotword detection plugin. Set this plugin to ``enabled`` if you
    want to run the hotword engine continuously instead of programmatically using
    ``start_detection`` and ``stop_detection``.

    Requires:

        - The :class:`platypush.plugins.stt.deepspeech.SttPicovoiceHotwordPlugin` plugin configured and its dependencies
          installed.

    """

    def __init__(self, *args, **kwargs):
        super().__init__('stt.picovoice.hotword', *args, **kwargs)


# vim:sw=4:ts=4:et:



import json

from platypush.backend import Backend
from platypush.message.event.scard import SmartCardDetectedEvent, SmartCardRemovedEvent


class ScardBackend(Backend):
    """
    Generic backend to read smart cards and NFC tags and trigger an event
    whenever a device is detected.

    Extend this backend to implement more advanced communication with custom
    smart cards.

    Triggers:

        * :class:`platypush.message.event.scard.SmartCardDetectedEvent` when a smart card is detected
        * :class:`platypush.message.event.scard.SmartCardRemovedEvent` when a smart card is removed

    Requires:

        * **pyscard** (``pip install pyscard``)
    """

    def __init__(self, atr=None, *args, **kwargs):
        """
        :param atr: If set, the backend will trigger events only for card(s) with the specified ATR(s). It can be either an ATR string (space-separated hex octects) or a list of ATR strings.  Default: none (any card will be detected)
        """

        from smartcard.CardType import AnyCardType, ATRCardType
        super().__init__(*args, **kwargs)
        self.ATRs = []

        if atr:
            if isinstance(atr, str):
                self.ATRs = [atr]
            elif isinstance(atr, list):
                self.ATRs = atr
            else:
                raise RuntimeError("Unsupported ATR: \"{}\" - type: {}, " +
                                   "supported types: string, list".format(
                                       atr, type(atr)))

            self.cardtype = ATRCardType( *[toBytes(atr) for atr in self.ATRs] )
        else:
            self.cardtype = AnyCardType()


    def run(self):
        from smartcard.CardRequest import CardRequest
        from smartcard.Exceptions import NoCardException, CardConnectionException
        from smartcard.util import toHexString

        super().run()

        self.logger.info('Initialized smart card reader backend - ATR filter: {}'.
                     format(self.ATRs))

        prev_atr = None
        reader = None

        while not self.should_stop():
            try:
                cardrequest = CardRequest(timeout=None, cardType=self.cardtype)
                cardservice = cardrequest.waitforcard()
                cardservice.connection.connect()

                reader = cardservice.connection.getReader()
                atr = toHexString(cardservice.connection.getATR())

                if atr != prev_atr:
                    self.logger.info('Smart card detected on reader {}, ATR: {}'.
                                format(reader, atr))

                    self.bus.post(SmartCardDetectedEvent(atr=atr, reader=reader))
                    prev_atr = atr
            except Exception as e:
                if isinstance(e, NoCardException) or isinstance(e, CardConnectionException):
                    self.bus.post(SmartCardRemovedEvent(atr=prev_atr, reader=reader))
                else:
                    self.logger.exception(e)

                prev_atr = None


# vim:sw=4:ts=4:et:




import re
import subprocess
import sys

from typing import Optional, List

from platypush.message.response.ping import PingResponse
from platypush.plugins import Plugin, action

PING_MATCHER = re.compile(
    r"(?P<min>\d+.\d+)/(?P<avg>\d+.\d+)/(?P<max>\d+.\d+)/(?P<mdev>\d+.\d+)"
)

PING_MATCHER_BUSYBOX = re.compile(
    r"(?P<min>\d+.\d+)/(?P<avg>\d+.\d+)/(?P<max>\d+.\d+)"
)

WIN32_PING_MATCHER = re.compile(r"(?P<min>\d+)ms.+(?P<max>\d+)ms.+(?P<avg>\d+)ms")


class PingPlugin(Plugin):
    """
    Perform ICMP network ping on remote hosts.
    """

    def __init__(self, executable: str = 'ping', count: int = 1, timeout: float = 5.0, **kwargs):
        """
        :param executable: Path to the ``ping`` executable. Default: the first ``ping`` executable found in PATH.
        :param count: Default number of packets that should be sent (default: 1).
        :param timeout: Default timeout before failing a ping request (default: 5 seconds).
        """

        super().__init__(**kwargs)
        self.executable = executable
        self.count = count
        self.timeout = timeout

    def _get_ping_cmd(self, host: str, count: int, timeout: float) -> List[str]:
        if sys.platform == 'win32':
            return [
                self.executable,
                '-n',
                str(count or self.count),
                '-w',
                str((timeout or self.timeout) * 1000),
                host,
            ]

        return [
            self.executable,
            '-n',
            '-q',
            '-c',
            str(count or self.count),
            '-W',
            str(timeout or self.timeout),
            host,
        ]

    @action
    def ping(self, host: str, count: Optional[int] = None, timeout: Optional[float] = None) -> PingResponse:
        """
        Ping a remote host.
        :param host: Remote host IP or name
        :param count: Number of packets that should be sent (default: 1).
        :param timeout: Timeout before failing a ping request (default: 5 seconds).
        """

        count = count or self.count
        timeout = timeout or self.timeout

        pinger = subprocess.Popen(
            self._get_ping_cmd(host, count=count, timeout=timeout),
            stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        try:
            out = pinger.communicate()
            if sys.platform == "win32":
                match = WIN32_PING_MATCHER.search(str(out).split("\n")[-1])
                min_val, avg_val, max_val = match.groups()
                mdev_val = None
            elif "max/" not in str(out):
                match = PING_MATCHER_BUSYBOX.search(str(out).split("\n")[-1])
                min_val, avg_val, max_val = match.groups()
                mdev_val = None
            else:
                match = PING_MATCHER.search(str(out).split("\n")[-1])
                min_val, avg_val, max_val, mdev_val = match.groups()

            return PingResponse(host=host, success=True, min=min_val, max=max_val, avg=avg_val, mdev=mdev_val)
        except (subprocess.CalledProcessError, AttributeError):
            return PingResponse(host=host, success=False)


# vim:sw=4:ts=4:et:



import re
import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.music import MusicPlayEvent, MusicPauseEvent, \
    MusicStopEvent, NewPlayingTrackEvent, PlaylistChangeEvent, VolumeChangeEvent, \
    PlaybackConsumeModeChangeEvent, PlaybackSingleModeChangeEvent, \
    PlaybackRepeatModeChangeEvent, PlaybackRandomModeChangeEvent


class MusicMpdBackend(Backend):
    """
    This backend listens for events on a MPD/Mopidy music server.

    Triggers:

        * :class:`platypush.message.event.music.MusicPlayEvent` if the playback state changed to play
        * :class:`platypush.message.event.music.MusicPauseEvent` if the playback state changed to pause
        * :class:`platypush.message.event.music.MusicStopEvent` if the playback state changed to stop
        * :class:`platypush.message.event.music.NewPlayingTrackEvent` if a new track is being played
        * :class:`platypush.message.event.music.PlaylistChangeEvent` if the main playlist has changed
        * :class:`platypush.message.event.music.VolumeChangeEvent` if the main volume has changed

    Requires:
        * **python-mpd2** (``pip install python-mpd2``)
        * The :mod:`platypush.plugins.music.mpd` plugin to be configured
    """

    def __init__(self, server='localhost', port=6600, poll_seconds=3, **kwargs):
        """
        :param poll_seconds: Interval between queries to the server (default: 3 seconds)
        :type poll_seconds: float
        """

        super().__init__(**kwargs)

        self.server = server
        self.port = port
        self.poll_seconds = poll_seconds


    def run(self):
        super().run()

        last_status = {}
        last_state = None
        last_track = None
        last_playlist = None
        plugin = None

        while not self.should_stop():
            success = False

            while not success:
                state = None
                playlist = None
                track = None

                try:
                    plugin = get_plugin('music.mpd')
                    if not plugin:
                        raise StopIteration

                    status = plugin.status().output
                    if not status or status.get('state') is None:
                        raise StopIteration

                    track = plugin.currentsong().output
                    state = status['state'].lower()
                    playlist = status['playlist']
                    success = True
                except Exception as e:
                    self.logger.debug(e)
                    get_plugin('music.mpd', reload=True)
                    if not state: state = last_state
                    if not playlist: playlist = last_playlist
                    if not track: track = last_track
                finally:
                    time.sleep(self.poll_seconds)

            if state != last_state:
                if state == 'stop':
                    self.bus.post(MusicStopEvent(status=status, track=track))
                elif state == 'pause':
                    self.bus.post(MusicPauseEvent(status=status, track=track))
                elif state == 'play':
                    self.bus.post(MusicPlayEvent(status=status, track=track))

            if playlist != last_playlist:
                if last_playlist:
                    # XXX plchanges can become heavy with big playlists,
                    # PlaylistChangeEvent temporarily disabled
                    # changes = plugin.plchanges(last_playlist).output
                    # self.bus.post(PlaylistChangeEvent(changes=changes))
                    self.bus.post(PlaylistChangeEvent())
                last_playlist = playlist

            if state == 'play' and track != last_track:
                self.bus.post(NewPlayingTrackEvent(status=status, track=track))

            if last_status.get('volume', None) != status['volume']:
                self.bus.post(VolumeChangeEvent(
                    volume=int(status['volume']), status=status, track=track))

            if last_status.get('random', None) != status['random']:
                self.bus.post(PlaybackRandomModeChangeEvent(
                    state=bool(int(status['random'])), status=status, track=track))

            if last_status.get('repeat', None) != status['repeat']:
                self.bus.post(PlaybackRepeatModeChangeEvent(
                    state=bool(int(status['repeat'])), status=status, track=track))

            if last_status.get('consume', None) != status['consume']:
                self.bus.post(PlaybackConsumeModeChangeEvent(
                    state=bool(int(status['consume'])), status=status, track=track))

            if last_status.get('single', None) != status['single']:
                self.bus.post(PlaybackSingleModeChangeEvent(
                    state=bool(int(status['single'])), status=status, track=track))

            last_status = status
            last_state = state
            last_track = track
            time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:




from platypush.message.event import Event


class SttEvent(Event):
    """ Base class for speech-to-text events """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class SpeechStartedEvent(SttEvent):
    """
    Event triggered when speech starts being detected.
    """
    pass


class SpeechDetectedEvent(SttEvent):
    """
    Event triggered when speech is detected.
    """

    def __init__(self, speech: str, *args, **kwargs):
        """
        :param speech: Speech detected, as a string
        """
        super().__init__(*args, speech=speech.strip(), **kwargs)


class ConversationDetectedEvent(SpeechDetectedEvent):
    """
    Event triggered when speech is detected after a hotword.
    """
    pass

class HotwordDetectedEvent(SttEvent):
    """
    Event triggered when a custom hotword is detected.
    """

    def __init__(self, hotword: str = '', *args, **kwargs):
        """
        :param hotword: The detected user hotword.
        """
        super().__init__(*args, hotword=hotword, **kwargs)


class SpeechDetectionStartedEvent(SttEvent):
    """
    Event triggered when the speech detection engine starts.
    """
    pass


class SpeechDetectionStoppedEvent(SttEvent):
    """
    Event triggered when the speech detection engine stops.
    """
    pass


# vim:sw=4:ts=4:et:



import datetime
import logging
import os
import re
import socket
import sys
import yaml

from platypush.utils import get_hash

""" Config singleton instance """
_default_config_instance = None

class Config(object):
    """
    Configuration base class
    Usage:
        - Initialize config from one of the default paths:
            Config.init()
        - Initialize config from a custom path
            Config.init(config_file_path)
        - Get a value
            Config.get('foo')
    """

    """
    Default config file locations:
        - $HOME/.config/platypush/config.yaml
        - /etc/platypush/config.yaml
    """
    _cfgfile_locations = [
        os.path.join(os.path.expanduser('~'), '.config', 'platypush', 'config.yaml'),
        os.path.join(os.sep, 'etc', 'platypush', 'config.yaml'),
    ]

    _default_constants = {
        'today': datetime.date.today,
        'now': datetime.datetime.now,
    }

    _workdir_location = os.path.join(os.path.expanduser('~'), '.local', 'share', 'platypush')
    _included_files = set()

    def __init__(self, cfgfile=None):
        """
        Constructor. Always use the class as a singleton (i.e. through
        Config.init), you won't probably need to call the constructor directly
        Params:
            cfgfile -- Config file path (default: retrieve the first
                       available location in _cfgfile_locations)
        """

        if cfgfile is None:
            cfgfile = self._get_default_cfgfile()

        if cfgfile is None:
            raise RuntimeError('No config file specified and nothing found in {}'
                               .format(self._cfgfile_locations))

        self._cfgfile = cfgfile
        self._config = self._read_config_file(self._cfgfile)

        if 'token' in self._config:
            self._config['token'] = self._config['token']
            self._config['token_hash'] = get_hash(self._config['token'])

        if 'workdir' not in self._config:
            self._config['workdir'] = self._workdir_location
        os.makedirs(self._config['workdir'], exist_ok=True)

        self._config['db'] = self._config.get('main.db', {
            'engine': 'sqlite:///' + os.path.join(
                os.path.expanduser('~'), '.local', 'share', 'platypush', 'main.db')
        })

        logging_config = {
            'level': logging.INFO,
            'stream': sys.stdout,
            'format': '%(asctime)-15s|%(levelname)5s|%(name)s|%(message)s',
        }

        if 'logging' in self._config:
            for (k,v) in self._config['logging'].items():
                if k == 'filename':
                    logfile = os.path.expanduser(v)
                    logdir = os.path.dirname(logfile)
                    try:
                        os.makedirs(logdir, exist_ok=True)
                    except Exception as e:
                        print('Unable to create logs directory {}: {}'.format(
                            logdir, str(e)))

                    v = logfile
                    del logging_config['stream']
                elif k == 'level':
                    try:
                        v = int(v)
                    except ValueError:
                        v = getattr(logging, v.upper())

                logging_config[k] = v

        self._config['logging'] = logging_config

        if 'device_id' not in self._config:
            self._config['device_id'] = socket.gethostname()

        if 'environment' in self._config:
            for k,v in self._config['environment'].items():
                os.environ[k] = str(v)

        self.backends = {}
        self.plugins = {}
        self.event_hooks = {}
        self.procedures = {}
        self.constants = {}
        self.cronjobs = {}

        self._init_constants()
        self._init_components()

    @staticmethod
    def _is_special_token(token):
        return token == 'main.db' or \
                token == 'token' or \
                token == 'token_hash' or \
                token == 'logging' or \
                token == 'workdir' or \
                token == 'device_id' or \
                token == 'environment'

    def _read_config_file(self, cfgfile):
        cfgfile_dir = os.path.dirname(os.path.abspath(
            os.path.expanduser(cfgfile)))

        config = {}

        with open(cfgfile, 'r') as fp:
            file_config = yaml.safe_load(fp)

        if not file_config:
            return config

        for section in file_config:
            if section == 'include':
                include_files = file_config[section] \
                    if isinstance(file_config[section], list) \
                    else [file_config[section]]

                for include_file in include_files:
                    if not os.path.isabs(include_file):
                        include_file = os.path.join(cfgfile_dir, include_file)
                    self._included_files.add(include_file)

                    included_config = self._read_config_file(include_file)
                    for incl_section in included_config.keys():
                        config[incl_section] = included_config[incl_section]
            elif 'disabled' not in file_config[section] \
                    or file_config[section]['disabled'] is False:
                config[section] = file_config[section]

        return config

    def _init_components(self):
        for key in self._config.keys():
            if key.startswith('backend.'):
                backend_name = '.'.join(key.split('.')[1:])
                self.backends[backend_name] = self._config[key]
            elif key.startswith('event.hook.'):
                hook_name = '.'.join(key.split('.')[2:])
                self.event_hooks[hook_name] = self._config[key]
            elif key.startswith('cron.'):
                cron_name = '.'.join(key.split('.')[1:])
                self.cronjobs[cron_name] = self._config[key]
            elif key.startswith('procedure.'):
                tokens = key.split('.')
                _async = True if len(tokens) > 2 and tokens[1] == 'async' else False
                procedure_name = '.'.join(tokens[2:] if len(tokens) > 2 else tokens[1:])
                args = []
                m = re.match(r'^([^(]+)\(([^)]+)\)\s*', procedure_name)

                if m:
                    procedure_name = m.group(1).strip()
                    args = [
                        arg.strip()
                        for arg in m.group(2).strip().split(',')
                        if arg.strip()
                    ]

                self.procedures[procedure_name] = {
                    '_async': _async,
                    'actions': self._config[key],
                    'args': args,
                }
            elif not self._is_special_token(key):
                self.plugins[key] = self._config[key]

    def _init_constants(self):
        if 'constants' in self._config:
            self.constants = self._config['constants']

        for (key,value) in self._default_constants.items():
            self.constants[key] = value


    @staticmethod
    def get_backends():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.backends

    @staticmethod
    def get_plugins():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.plugins

    @staticmethod
    def get_event_hooks():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.event_hooks

    @staticmethod
    def get_procedures():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.procedures

    @staticmethod
    def get_constants():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        constants = {}

        for name in _default_config_instance.constants.keys():
            constants[name] = Config.get_constant(name)
        return constants

    @staticmethod
    def get_constant(name):
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()

        if name not in _default_config_instance.constants: return None
        value = _default_config_instance.constants[name]
        return value() if callable(value) else value

    @staticmethod
    def get_cronjobs():
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance.cronjobs

    @staticmethod
    def get_default_pusher_backend():
        """
        Gets the default pusher backend from the config
        """
        backends = [k for k in Config.get_backends().keys()
                    if 'pusher' in Config.get_backends()[k]
                    and Config.get_backends()[k]['pusher'] is True]

        return backends[0] if backends else None


    @classmethod
    def _get_default_cfgfile(cls):
        for location in cls._cfgfile_locations:
            if os.path.isfile(location): return location

    @staticmethod
    def init(cfgfile=None):
        """
        Initializes the config object singleton
        Params:
            cfgfile -- path to the config file - default: _cfgfile_locations
        """
        global _default_config_instance
        _default_config_instance = Config(cfgfile)

    @staticmethod
    def get(key):
        """
        Gets a config value
        Params:
            key -- Config key to get
        """
        global _default_config_instance
        if _default_config_instance is None: _default_config_instance = Config()
        return _default_config_instance._config.get(key)


# vim:sw=4:ts=4:et:




import time

from platypush.backend import Backend
from platypush.context import get_plugin
from platypush.message.event.sensor import SensorDataChangeEvent, \
    SensorDataAboveThresholdEvent, SensorDataBelowThresholdEvent


class SensorBackend(Backend):
    """
    Abstract backend for polling sensors.

    Triggers:

        * :class:`platypush.message.event.sensor.SensorDataChangeEvent` if the measurements of a sensor have changed
        * :class:`platypush.message.event.sensor.SensorDataAboveThresholdEvent` if the measurements of a sensor have
            gone above a configured threshold
        * :class:`platypush.message.event.sensor.SensorDataBelowThresholdEvent` if the measurements of a sensor have
            gone below a configured threshold
    """

    default_tolerance = 1e-7

    def __init__(self, plugin=None, plugin_args=None, thresholds=None, tolerance=default_tolerance, poll_seconds=None,
                 enabled_sensors=None, **kwargs):
        """
        :param plugin: If set, then this plugin instance, referenced by plugin id, will be polled
            through ``get_plugin()``. Example: ``'gpio.sensor.bme280'`` or ``'gpio.sensor.envirophat'``.
        :type plugin: str

        :param plugin_args: If plugin is set and its ``get_measurement()`` method accepts optional arguments, then you
            can pass those arguments through ``plugin_args``.
        :type plugin_args: dict

        :param thresholds: Thresholds can be either a scalar value or a dictionary (e.g. ``{"temperature": 20.0}``).
            Sensor threshold events will be fired when measurements get above or below these values.
            Set it as a scalar if your get_measurement() code returns a scalar, as a dictionary if it returns a
            dictionary of values.  For instance, if your sensor code returns both humidity and temperature in a format
            like ``{'humidity':60.0, 'temperature': 25.0}``, you'll want to set up a threshold on temperature with a
            syntax like ``{'temperature':20.0}`` to trigger events when the temperature goes above/below 20 degrees.

        :param tolerance: If set, then the sensor change events will be triggered only if the difference between
            the new value and the previous value is higher than the specified tolerance. Example::

                {
                    "temperature": 0.01,  # Tolerance on the 2nd decimal digit
                    "humidity": 0.1       # Tolerance on the 1st decimal digit
                }

        :type tolerance: dict or float

        :param poll_seconds: If set, the thread will wait for the specified number of seconds between a read and the
            next one.
        :type poll_seconds: float

        :param enabled_sensors: If ``get_measurement()`` returns data in dict form, then ``enabled_sensors`` selects
            which keys should be taken into account when monitoring for new events (e.g. "temperature" or "humidity").
        :type enabled_sensors: dict (in the form ``name -> [True/False]``), set or list
        """

        super().__init__(**kwargs)

        self.data = None
        self.plugin = plugin
        self.plugin_args = plugin_args or {}
        self.thresholds = thresholds
        self.tolerance = tolerance
        self.poll_seconds = poll_seconds

        if isinstance(enabled_sensors, list):
            enabled_sensors = set(enabled_sensors)
        if isinstance(enabled_sensors, set):
            enabled_sensors = {k: True for k in enabled_sensors}

        self.enabled_sensors = enabled_sensors or {}

    def get_measurement(self):
        """
        Wrapper around ``plugin.get_measurement()`` that can filter events on specified enabled sensors data or on
        specified tolerance values. It can be overridden by derived classes.
        """
        if not self.plugin:
            raise NotImplementedError('No plugin specified')

        reload = False
        success = False
        data = None

        while not success:
            try:
                plugin = get_plugin(self.plugin, reload=reload)
                data = plugin.get_data(**self.plugin_args).output
                if reload:
                    self.logger.info('Backend successfully restored')

                success = True
            except Exception as e:
                self.logger.warning('Unexpected exception while getting data: {}'.format(str(e)))
                self.logger.exception(e)
                reload = True
                time.sleep(5)

            if self.enabled_sensors and data is not None:
                data = {
                    sensor: data[sensor]
                    for sensor, enabled in self.enabled_sensors.items()
                    if enabled and sensor in data
                }

        return data

    @staticmethod
    def _get_value(value):
        if isinstance(value, float) or isinstance(value, int) or isinstance(value, bool):
            return value
        return float(value)

    def get_new_data(self, new_data):
        if self.data is None or new_data is None:
            return new_data

        # noinspection PyBroadException
        try:
            # Scalar data case
            new_data = self._get_value(new_data)
            return new_data if abs(new_data - self.data) >= self.tolerance else None
        except:
            # If it's not a scalar then it should be a dict
            assert isinstance(new_data, dict), 'Invalid type {} received for sensor data'.format(type(new_data))

        ret = {}
        for k, v in new_data.items():
            if (v is None and self.data.get(k) is not None) \
                    or k not in self.data \
                    or self.tolerance is None:
                ret[k] = v
                continue

            if v is None:
                continue

            tolerance = None
            is_nan = False
            old_v = None

            try:
                v = self._get_value(v)
                old_v = self._get_value(self.data.get(k))
            except (TypeError, ValueError):
                is_nan = True

            if not is_nan:
                if isinstance(self.tolerance, dict):
                    tolerance = float(self.tolerance.get(k, self.default_tolerance))
                else:
                    try:
                        tolerance = float(self.tolerance)
                    except (TypeError, ValueError):
                        pass

                if tolerance is None or abs(v - old_v) >= tolerance:
                    ret[k] = v
            elif k not in self.data or self.data[k] != v:
                ret[k] = v

        return ret

    def on_stop(self):
        if not self.plugin:
            return

        plugin = get_plugin(self.plugin)
        if plugin and hasattr(plugin, 'close'):
            plugin.close()

    def process_data(self, data, new_data):
        if new_data:
            self.bus.post(SensorDataChangeEvent(data=new_data, source=self.plugin or self.__class__.__name__))

    def run(self):
        super().run()
        self.logger.info('Initialized {} sensor backend'.format(self.__class__.__name__))

        while not self.should_stop():
            try:
                data = self.get_measurement()
                new_data = self.get_new_data(data)
                self.process_data(data, new_data)

                data_below_threshold = {}
                data_above_threshold = {}

                if self.thresholds:
                    if isinstance(self.thresholds, dict) and isinstance(data, dict):
                        for (measure, thresholds) in self.thresholds.items():
                            if measure not in data:
                                continue

                            if not isinstance(thresholds, list):
                                thresholds = [thresholds]

                            for threshold in thresholds:
                                if data[measure] > threshold and (self.data is None or (
                                        measure in self.data and self.data[measure] <= threshold)):
                                    data_above_threshold[measure] = data[measure]
                                elif data[measure] < threshold and (self.data is None or (
                                        measure in self.data and self.data[measure] >= threshold)):
                                    data_below_threshold[measure] = data[measure]

                if data_below_threshold:
                    self.bus.post(SensorDataBelowThresholdEvent(data=data_below_threshold))

                if data_above_threshold:
                    self.bus.post(SensorDataAboveThresholdEvent(data=data_above_threshold))

                self.data = data

                if new_data:
                    if isinstance(new_data, dict):
                        for k, v in new_data.items():
                            self.data[k] = v
                    else:
                        self.data = new_data
            except Exception as e:
                self.logger.exception(e)

            if self.poll_seconds:
                time.sleep(self.poll_seconds)


# vim:sw=4:ts=4:et:



import os
import select
import subprocess
import threading
import time

from platypush.context import get_bus
from platypush.message.response import Response
from platypush.plugins.media import PlayerState, MediaPlugin
from platypush.message.event.media import MediaPlayEvent, MediaPlayRequestEvent, \
    MediaPauseEvent, MediaStopEvent, NewPlayingMediaEvent

from platypush.plugins import action
from platypush.utils import find_bins_in_path


class MediaMplayerPlugin(MediaPlugin):
    """
    Plugin to control MPlayer instances

    Requires:

        * **mplayer** executable on your system
    """

    _mplayer_default_communicate_timeout = 0.5

    _mplayer_bin_default_args = ['-slave', '-quiet', '-idle', '-input',
                                 'nodefault-bindings', '-noconfig', 'all']

    def __init__(self, mplayer_bin=None,
                 mplayer_timeout=_mplayer_default_communicate_timeout,
                 args=None, *argv, **kwargs):
        """
        Create the MPlayer wrapper. Note that the plugin methods are populated
        dynamically by introspecting the mplayer executable. You can verify the
        supported methods at runtime by using the `list_actions` method.

        :param mplayer_bin: Path to the MPlayer executable (default: search for
            the first occurrence in your system PATH environment variable)
        :type mplayer_bin: str

        :param mplayer_timeout: Timeout in seconds to wait for more data
            from MPlayer before considering a response ready (default: 0.5 seconds)
        :type mplayer_timeout: float

        :param subtitles: Path to the subtitles file
        :type subtitles: str

        :param args: Default arguments that will be passed to the MPlayer
            executable
        :type args: list
        """

        super().__init__(*argv, **kwargs)

        self.args = args or []
        self._init_mplayer_bin(mplayer_bin=mplayer_bin)
        self._build_actions()
        self._player = None
        self._mplayer_timeout = mplayer_timeout
        self._mplayer_stopped_event = threading.Event()
        self._status_lock = threading.Lock()

    def _init_mplayer_bin(self, mplayer_bin=None):
        if not mplayer_bin:
            bin_name = 'mplayer.exe' if os.name == 'nt' else 'mplayer'
            bins = find_bins_in_path(bin_name)

            if not bins:
                raise RuntimeError('MPlayer executable not specified and not ' +
                                   'found in your PATH. Make sure that mplayer' +
                                   'is either installed or configured')

            self.mplayer_bin = bins[0]
        else:
            mplayer_bin = os.path.expanduser(mplayer_bin)
            if not (os.path.isfile(mplayer_bin)
                    and (os.name == 'nt' or os.access(mplayer_bin, os.X_OK))):
                raise RuntimeError('{} is does not exist or is not a valid ' +
                                   'executable file'.format(mplayer_bin))

            self.mplayer_bin = mplayer_bin

    def _init_mplayer(self, mplayer_args=None):
        if self._player:
            try:
                self._player.terminate()
            except Exception as e:
                self.logger.debug('Failed to quit mplayer before _exec: {}'.
                                  format(str(e)))

        mplayer_args = mplayer_args or []
        args = [self.mplayer_bin] + self._mplayer_bin_default_args
        for arg in self.args + mplayer_args:
            if arg not in args:
                args.append(arg)

        popen_args = {
            'stdin': subprocess.PIPE,
            'stdout': subprocess.PIPE,
        }

        if self._env:
            popen_args['env'] = self._env

        self._player = subprocess.Popen(args, **popen_args)
        threading.Thread(target=self._process_monitor()).start()

    def _build_actions(self):
        """ Populates the actions list by introspecting the mplayer executable """

        self._actions = {}
        mplayer = subprocess.Popen([self.mplayer_bin, '-input', 'cmdlist'],
                                   stdin=subprocess.PIPE, stdout=subprocess.PIPE)

        def args_pprint(txt):
            lc = txt.lower()
            if lc[0] == '[':
                return '%s=None' % lc[1:-1]
            return lc

        while True:
            line = mplayer.stdout.readline()
            if not line:
                break
            line = line.decode()
            if line[0].isupper():
                continue
            args = line.split()
            cmd_name = args.pop(0)
            arguments = ', '.join([args_pprint(a) for a in args])
            self._actions[cmd_name] = '{}({})'.format(cmd_name, arguments)

    def _exec(self, cmd, *args, mplayer_args=None, prefix=None,
              wait_for_response=False):
        cmd_name = cmd
        response = None

        if cmd_name == 'loadfile' or cmd_name == 'loadlist':
            self._init_mplayer(mplayer_args)
        else:
            if not self._player:
                self.logger.warning('MPlayer is not running')

        cmd = '{}{}{}{}\n'.format(
            prefix + ' ' if prefix else '',
            cmd_name, ' ' if args else '',
            ' '.join(repr(a) for a in args)).encode()

        if not self._player:
            self.logger.warning('Cannot send command {}: player unavailable'.format(cmd))
            return

        self._player.stdin.write(cmd)
        self._player.stdin.flush()

        if cmd_name == 'loadfile' or cmd_name == 'loadlist':
            self._post_event(NewPlayingMediaEvent, resource=args[0])
        elif cmd_name == 'pause':
            self._post_event(MediaPauseEvent)
        elif cmd_name == 'quit' or cmd_name == 'stop':
            if cmd_name == 'quit':
                self._player.terminate()
                self._player.wait()
                try:
                    self._player.kill()
                except Exception:
                    pass
                self._player = None

        if not wait_for_response:
            return

        poll = select.poll()
        poll.register(self._player.stdout, select.POLLIN)
        last_read_time = time.time()

        while time.time() - last_read_time < self._mplayer_timeout:
            result = poll.poll(0)
            if result:
                if not self._player:
                    break

                line = self._player.stdout.readline().decode()
                last_read_time = time.time()

                if line.startswith('ANS_'):
                    k, v = tuple(line[4:].split('='))
                    v = v.strip()
                    if v == 'yes':
                        v = True
                    elif v == 'no':
                        v = False

                    try:
                        v = eval(v)
                    except Exception:
                        pass

                    response = {k: v}

        return response

    @action
    def execute(self, cmd, args=None):
        """
        Execute a raw MPlayer command. See
        http://www.mplayerhq.hu/DOCS/tech/slave.txt for a reference or call
        :meth:`platypush.plugins.media.mplayer.list_actions` to get a list
        """

        args = args or []
        return self._exec(cmd, *args)

    @action
    def list_actions(self):
        return [{'action': a, 'args': self._actions[a]}
                for a in sorted(self._actions.keys())]

    def _process_monitor(self):
        def _thread():
            if not self._player:
                return

            self._mplayer_stopped_event.clear()
            self._player.wait()
            try:
                self.quit()
            except Exception:
                pass

            self._post_event(MediaStopEvent)
            self._mplayer_stopped_event.set()
            self._player = None

        return _thread

    @staticmethod
    def _post_event(evt_type, **evt):
        bus = get_bus()
        bus.post(evt_type(player='local', plugin='media.mplayer', **evt))

    @action
    def play(self, resource, subtitles=None, mplayer_args=None):
        """
        Play a resource.

        :param resource: Resource to play - can be a local file or a remote URL
        :type resource: str

        :param subtitles: Path to optional subtitle file
        :type subtitles: str

        :param mplayer_args: Extra runtime arguments that will be passed to the
            MPlayer executable
        :type mplayer_args: list[str]
        """

        self._post_event(MediaPlayRequestEvent, resource=resource)
        if subtitles:
            mplayer_args = mplayer_args or []
            mplayer_args += ['-sub', self.get_subtitles_file(subtitles)]

        resource = self._get_resource(resource)
        if resource.startswith('file://'):
            resource = resource[7:]

        self._exec('loadfile', resource, mplayer_args=mplayer_args)
        self._post_event(MediaPlayEvent, resource=resource)

        if self.volume:
            self.set_volume(volume=self.volume)

        return self.status()

    @action
    def pause(self):
        """ Toggle the paused state """
        self._exec('pause')
        self._post_event(MediaPauseEvent)
        return self.status()

    @action
    def stop(self):
        """ Stop the playback """
        # return self._exec('stop')
        self.quit()
        return self.status()

    @action
    def quit(self):
        """ Quit the player """
        self._stop_torrent()
        self._exec('quit')
        self._post_event(MediaStopEvent)
        return self.status()

    @action
    def voldown(self, step=10.0):
        """ Volume down by (default: 10)% """
        self._exec('volume', -step * 10)
        return self.status()

    @action
    def volup(self, step=10.0):
        """ Volume up by (default: 10)% """
        self._exec('volume', step * 10)
        return self.status()

    @action
    def back(self, offset=60.0):
        """ Back by (default: 60) seconds """
        self.step_property('time_pos', -offset)
        return self.status()

    @action
    def forward(self, offset=60.0):
        """ Forward by (default: 60) seconds """
        self.step_property('time_pos', offset)
        return self.status()

    @action
    def toggle_subtitles(self):
        """ Toggle the subtitles visibility """
        subs = self.get_property('sub_visibility').output.get('sub_visibility')
        self._exec('sub_visibility', int(not subs))
        return self.status()

    @action
    def add_subtitles(self, filename, **args):
        """ Sets media subtitles from filename """
        self._exec('sub_visibility', 1)
        self._exec('sub_load', filename)
        return self.status()

    @action
    def remove_subtitles(self, index=None):
        """ Removes the subtitle specified by the index (default: all) """
        if index is None:
            self._exec('sub_remove')
        else:
            self._exec('sub_remove', index)

        return self.status()

    @action
    def is_playing(self):
        """
        :returns: True if it's playing, False otherwise
        """
        return self.get_property('pause').output.get('pause') is False

    @action
    def load(self, resource, mplayer_args=None, **kwargs):
        """
        Load a resource/video in the player.
        """
        if mplayer_args is None:
            mplayer_args = {}
        return self.play(resource, mplayer_args=mplayer_args)

    @action
    def mute(self):
        """ Toggle mute state """
        self._exec('mute')
        return self.status()

    @action
    def seek(self, position):
        """
        Seek backward/forward by the specified number of seconds

        :param position: Number of seconds relative to the current cursor
        :type position: int
        """
        self.step_property('time_pos', position)
        return self.status()

    @action
    def set_position(self, position):
        """
        Seek backward/forward to the specified absolute position

        :param position: Number of seconds from the start
        :type position: int
        """
        self.set_property('time_pos', position)
        return self.status()

    @action
    def set_volume(self, volume):
        """
        Set the volume

        :param volume: Volume value between 0 and 100
        :type volume: float
        """
        self._exec('volume', volume)
        return self.status()

    @action
    def status(self):
        """
        Get the current player state.

        :returns: A dictionary containing the current state.

        Example::

            output = {
                "state": "play"  # or "stop" or "pause"
            }
        """

        if not self._player:
            return {'state': PlayerState.STOP.value}

        status = {}
        props = {
            'duration': 'length',
            'filename': 'filename',
            'fullscreen': 'fullscreen',
            'mute': 'mute',
            'name': 'filename',
            'path': 'path',
            'pause': 'pause',
            'percent_pos': 'percent_pos',
            'position': 'time_pos',
            'title': 'filename',
            'volume': 'volume',
        }

        with self._status_lock:
            for prop, player_prop in props.items():
                value = self.get_property(player_prop).output
                if value is not None:
                    status[prop] = value.get(player_prop)

        status['seekable'] = True if status['duration'] is not None else False
        status['state'] = PlayerState.PAUSE.value if status['pause'] else PlayerState.PLAY.value

        if status['path']:
            status['url'] = ('file://' if os.path.isfile(status['path']) else '') + status['path']

        status['volume_max'] = 100
        return status

    @action
    def get_property(self, property, args=None):
        """
        Get a player property (e.g. pause, fullscreen etc.). See
        http://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available properties
        """

        args = args or []
        response = Response(output={})

        result = self._exec('get_property', property, prefix='pausing_keep_force',
                            wait_for_response=True, *args) or {}

        for k, v in result.items():
            if k == 'ERROR' and v not in response.errors:
                response.errors.append('{}{}: {}'.format(property, args, v))
            else:
                response.output[k] = v

        return response

    @action
    def set_property(self, property, value, args=None):
        """
        Set a player property (e.g. pause, fullscreen etc.). See
        http://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available properties
        """

        args = args or []
        response = Response(output={})

        result = self._exec('set_property', property, value,
                            prefix='pausing_keep_force' if property != 'pause'
                            else None, wait_for_response=True, *args) or {}

        for k, v in result.items():
            if k == 'ERROR' and v not in response.errors:
                response.errors.append('{} {}{}: {}'.format(property, value,
                                                            args, v))
            else:
                response.output[k] = v

        return response

    @action
    def step_property(self, property, value, args=None):
        """
        Step a player property (e.g. volume, time_pos etc.). See
        http://www.mplayerhq.hu/DOCS/tech/slave.txt for a full list of the
        available steppable properties
        """

        args = args or []
        response = Response(output={})

        result = self._exec('step_property', property, value,
                            prefix='pausing_keep_force',
                            wait_for_response=True, *args) or {}

        for k, v in result.items():
            if k == 'ERROR' and v not in response.errors:
                response.errors.append('{} {}{}: {}'.format(property, value,
                                                            args, v))
            else:
                response.output[k] = v

        return response

# vim:sw=4:ts=4:et:



import os
import re

from flask import Blueprint, abort, send_from_directory

from platypush.config import Config
from platypush.backend.http.app import template_folder, static_folder


img_folder = os.path.join(static_folder, 'resources', 'img')
resources = Blueprint('resources', __name__, template_folder=template_folder)
favicon = Blueprint('favicon', __name__, template_folder=template_folder)
img = Blueprint('img', __name__, template_folder=template_folder)

# Declare routes list
__routes__ = [
    resources,
    favicon,
    img,
]


@resources.route('/resources/<path:path>', methods=['GET'])
def resources_path(path):
    """ Custom static resources """
    path_tokens = path.split('/')
    filename = path_tokens.pop(-1)
    http_conf = Config.get('backend.http')
    resource_dirs = http_conf.get('resource_dirs', {})

    while path_tokens:
        if '/'.join(path_tokens) in resource_dirs:
            break
        path_tokens.pop()

    if not path_tokens:
        # Requested resource not found in the allowed resource_dirs
        abort(404)

    base_path = '/'.join(path_tokens)
    real_base_path = os.path.abspath(os.path.expanduser(resource_dirs[base_path]))
    real_path = real_base_path

    file_path = [s for s in re.sub(r'^{}(.*)$'.format(base_path), '\\1', path)
                 .split('/') if s]

    for p in file_path[:-1]:
        real_path += os.sep + p
        file_path.pop(0)

    file_path = file_path.pop(0)
    if not real_path.startswith(real_base_path):
        # Directory climbing attempt
        abort(404)

    return send_from_directory(real_path, file_path)


@favicon.route('/favicon.ico', methods=['GET'])
def favicon():
    """ favicon.ico icon """
    return send_from_directory(img_folder, 'favicon.ico')

@img.route('/img/<path:path>', methods=['GET'])
def imgpath(path):
    """ Default static images """
    return send_from_directory(img_folder, path)


# vim:sw=4:ts=4:et:



from abc import ABC, abstractmethod

from platypush.context import get_backend
from platypush.plugins import Plugin, action


class AssistantPlugin(ABC, Plugin):
    """
    Base class for assistant plugins
    """

    @abstractmethod
    def start_conversation(self, *args, language=None, tts_plugin=None, tts_args=None, **kwargs):
        """
        Start a conversation.
        """
        raise NotImplementedError

    @abstractmethod
    def stop_conversation(self, *args, **kwargs):
        """
        Stop a conversation.
        """
        raise NotImplementedError

    def _get_assistant(self):
        return get_backend('assistant.snowboy')

    @action
    def pause_detection(self):
        """
        Put the assistant on pause. No new conversation events will be triggered.
        """
        assistant = self._get_assistant()
        assistant.pause_detection()

    @action
    def resume_detection(self):
        """
        Resume the assistant hotword detection from a paused state.
        """
        assistant = self._get_assistant()
        assistant.resume_detection()

    @action
    def is_detecting(self) -> bool:
        """
        :return: True if the asistant is detecting, False otherwise.
        """
        assistant = self._get_assistant()
        return assistant.is_detecting()


# vim:sw=4:ts=4:et:



from typing import Optional

from platypush.message.event import Event


class SensorDataChangeEvent(Event):
    """
    Event triggered when a sensor has new data
    """

    def __init__(self, data, source: Optional[str] = None, *args, **kwargs):
        """
        :param data: Sensor data
        """

        super().__init__(data=data, source=source, *args, **kwargs)
        self.data = data
        self.source = source


class SensorDataAboveThresholdEvent(Event):
    """
    Event triggered when a sensor's read goes above a configured threshold
    """

    def __init__(self, data, *args, **kwargs):
        """
        :param data: Sensor data
        """

        super().__init__(data=data, *args, **kwargs)
        self.data = data


class SensorDataBelowThresholdEvent(Event):
    """
    Event triggered when a sensor's read goes below a configured threshold
    """

    def __init__(self, data, *args, **kwargs):
        """
        :param data: Sensor data
        """

        super().__init__(data=data, *args, **kwargs)
        self.data = data


# vim:sw=4:ts=4:et:



import datetime
import os
import re
import time

from sqlalchemy import create_engine, func, or_, Column, Integer, String, \
    DateTime, PrimaryKeyConstraint, ForeignKey

from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.sql.expression import func

from platypush.config import Config
from platypush.plugins.media import MediaPlugin
from platypush.plugins.media.search import MediaSearcher

Base = declarative_base()
Session = scoped_session(sessionmaker())

class LocalMediaSearcher(MediaSearcher):
    """
    This class will search for media in the local configured directories. It
    will index the media files for a faster search, it will detect which
    directories have been changed since the last scan and re-index their content
    if needed.

    Requires:

        * **sqlalchemy** (``pip install sqlalchemy``)
    """

    _filename_separators = '[.,_\-@()\[\]\{\}\s\'\"]+'

    def __init__(self, dirs, *args, **kwargs):
        super().__init__()
        self.dirs = dirs
        db_dir = os.path.join(Config.get('workdir'), 'media')
        os.makedirs(db_dir, exist_ok=True)
        self.db_file = os.path.join(db_dir, 'media.db')
        self._db_engine = None

    def _get_db_session(self):
        if not self._db_engine:
            self._db_engine = create_engine(
                'sqlite:///{}'.format(self.db_file),
                connect_args = {'check_same_thread': False})

            Base.metadata.create_all(self._db_engine)
            Session.configure(bind=self._db_engine)

        return Session()

    @staticmethod
    def _get_or_create_dir_entry(session, path):
        record = session.query(MediaDirectory).filter_by(path=path).first()
        if record is None:
            record = MediaDirectory.build(path=path)
            session.add(record)

        session.commit()
        return record

    @classmethod
    def _get_last_modify_time(cls, path, recursive=False):
        return max([os.path.getmtime(p) for p,_,_ in os.walk(path)]) \
            if recursive else os.path.getmtime(path)

    @classmethod
    def _has_directory_changed_since_last_indexing(self, dir_record):
        if not dir_record.last_indexed_at:
            return True

        return datetime.datetime.fromtimestamp(
            self._get_last_modify_time(dir_record.path)) > dir_record.last_indexed_at

    @classmethod
    def _matches_query(cls, filename, query):
        filename = filename.lower()
        query_tokens = [_.lower() for _ in re.split(
            cls._filename_separators, query.strip())]

        for token in query_tokens:
            if token not in filename:
                return False
        return True

    @classmethod
    def _sync_token_records(cls, session, *tokens):
        tokens = list(tokens)
        if not tokens:
            return []

        records = { record.token: record for record in
                   session.query(MediaToken).filter(
                       MediaToken.token.in_(tokens)).all() }

        for token in tokens:
            if token in records:
                continue
            record = MediaToken.build(token=token)
            session.add(record)
            records[token] = record

        session.commit()
        return session.query(MediaToken).filter(
            MediaToken.token.in_(tokens)).all()

    @classmethod
    def _get_file_records(cls, dir_record, session):
        return session.query(MediaFile).filter_by(
            directory_id=dir_record.id).all()

    def scan(self, media_dir, session=None, dir_record=None):
        """
        Scans a media directory and stores the search results in the internal
        SQLite index
        """

        if not session:
            session = self._get_db_session()

        self.logger.info('Indexing directory {}'.format(media_dir))
        index_start_time = time.time()

        if not dir_record:
            dir_record = self._get_or_create_dir_entry(session, media_dir)

        if not os.path.isdir(media_dir):
            self.logger.info('Directory {} is no longer accessible, removing it'.
                             format(media_dir))
            session.query(MediaDirectory) \
                .filter(MediaDirectory.path==media_dir) \
                .delete(synchronize_session='fetch')
            return

        stored_file_records = {
            f.path: f for f in self._get_file_records(dir_record, session) }

        for path, dirs, files in os.walk(media_dir):
            for filename in files:
                filepath = os.path.join(path, filename)

                if filepath in stored_file_records:
                    # stored_file_records will be used to keep track of the
                    # files that have been removed from media_dir. If the file
                    # is still there, don't index it again
                    del stored_file_records[filepath]
                    continue

                if not MediaPlugin._is_video_file(filename) and \
                        not MediaPlugin._is_audio_file(filename):
                    continue

                self.logger.debug('Syncing item {}'.format(filepath))
                tokens = [_.lower() for _ in re.split(self._filename_separators,
                                                      filename.strip())]

                token_records = self._sync_token_records(session, *tokens)
                file_record = MediaFile.build(directory_id=dir_record.id,
                                              path=filepath)

                session.add(file_record)
                session.commit()
                file_record = session.query(MediaFile).filter_by(
                    directory_id=dir_record.id, path=filepath).one()

                for token_record in token_records:
                    file_token = MediaFileToken.build(file_id=file_record.id,
                                                      token_id=token_record.id)
                    session.add(file_token)

        # stored_file_records should now only contain the records of the files
        # that have been removed from the directory
        if stored_file_records:
            self.logger.info(
                'Removing references to {} deleted media items from {}'.format(
                    len(stored_file_records), media_dir))

            session.query(MediaFile).filter(MediaFile.id.in_(
                [ record.id for record in stored_file_records.values() ]
            )).delete(synchronize_session='fetch')

        dir_record.last_indexed_at = datetime.datetime.now()
        self.logger.info('Scanned {} in {} seconds'.format(
            media_dir, int(time.time()-index_start_time)))

        session.commit()

    def search(self, query, **kwargs):
        """
        Searches in the configured media directories given a query. It uses the
        built-in SQLite index if available. If any directory has changed since
        the last scan then it will be indexed again and the up-to-date results
        will be returned.
        """

        session = self._get_db_session()
        results = {}

        for media_dir in self.dirs:
            self.logger.info('Searching {} for "{}"'.format(media_dir, query))
            dir_record = self._get_or_create_dir_entry(session, media_dir)

            if self._has_directory_changed_since_last_indexing(dir_record):
                self.logger.info('{} has changed since last indexing, '.format(
                    media_dir) + 're-indexing')

                self.scan(media_dir, session=session, dir_record=dir_record)

            query_tokens = [_.lower() for _ in re.split(
                self._filename_separators, query.strip())]

            for file_record in session.query(MediaFile.path). \
                    join(MediaFileToken). \
                    join(MediaToken). \
                    filter(MediaToken.token.in_(query_tokens)). \
                    group_by(MediaFile.path). \
                    having(func.count(MediaFileToken.token_id) >= len(query_tokens)):
                if os.path.isfile(file_record.path):
                    results[file_record.path] = {
                        'url': 'file://' + file_record.path,
                        'title': os.path.basename(file_record.path),
                        'size': os.path.getsize(file_record.path)
                    }

        return results.values()


# --- Table definitions

class MediaDirectory(Base):
    """ Models the MediaDirectory table """

    __tablename__ = 'MediaDirectory'
    __table_args__ = ({ 'sqlite_autoincrement': True })

    id = Column(Integer, primary_key=True)
    path = Column(String)
    last_indexed_at = Column(DateTime)

    @classmethod
    def build(cls, path, last_indexed_at=None, id=None):
        record = cls()
        record.id = id
        record.path = path
        record.last_indexed_at = last_indexed_at
        return record


class MediaFile(Base):
    """ Models the MediaFile table """

    __tablename__ = 'MediaFile'
    __table_args__ = ({ 'sqlite_autoincrement': True })

    id = Column(Integer, primary_key=True)
    directory_id = Column(Integer, ForeignKey(
        'MediaDirectory.id', ondelete='CASCADE'), nullable=False)
    path = Column(String, nullable=False, unique=True)
    indexed_at = Column(DateTime)

    @classmethod
    def build(cls, directory_id, path, indexed_at=None, id=None):
        record = cls()
        record.id = id
        record.directory_id = directory_id
        record.path = path
        record.indexed_at = indexed_at or datetime.datetime.now()
        return record


class MediaToken(Base):
    """ Models the MediaToken table """

    __tablename__ = 'MediaToken'
    __table_args__ = ({ 'sqlite_autoincrement': True })

    id = Column(Integer, primary_key=True)
    token = Column(String, nullable=False, unique=True)

    @classmethod
    def build(cls, token, id=None):
        record = cls()
        record.id = id
        record.token = token
        return record


class MediaFileToken(Base):
    """ Models the MediaFileToken table """

    __tablename__ = 'MediaFileToken'

    file_id = Column(Integer, ForeignKey('MediaFile.id', ondelete='CASCADE'),
                     nullable=False)
    token_id = Column(Integer, ForeignKey('MediaToken.id', ondelete='CASCADE'),
                      nullable=False)

    __table_args__ = ( PrimaryKeyConstraint(file_id, token_id), {} )

    @classmethod
    def build(cls, file_id, token_id, id=None):
        record = cls()
        record.id = id
        record.file_id = file_id
        record.token_id = token_id
        return record


# vim:sw=4:ts=4:et:



import json
import socket
import threading
import time

from platypush.backend import Backend
from platypush.utils import set_thread_name
from platypush.message.event.music.snapcast import ClientVolumeChangeEvent, \
    GroupMuteChangeEvent, ClientConnectedEvent, ClientDisconnectedEvent, \
    ClientLatencyChangeEvent, ClientNameChangeEvent, GroupStreamChangeEvent, \
    StreamUpdateEvent, ServerUpdateEvent


class MusicSnapcastBackend(Backend):
    """
    Backend that listens for notification and status changes on one or more
    [Snapcast](https://github.com/badaix/snapcast) servers.

    Triggers:

        * :class:`platypush.message.event.music.snapcast.ClientConnectedEvent`
        * :class:`platypush.message.event.music.snapcast.ClientDisconnectedEvent`
        * :class:`platypush.message.event.music.snapcast.ClientVolumeChangeEvent`
        * :class:`platypush.message.event.music.snapcast.ClientLatencyChangeEvent`
        * :class:`platypush.message.event.music.snapcast.ClientNameChangeEvent`
        * :class:`platypush.message.event.music.snapcast.GroupMuteChangeEvent`
        * :class:`platypush.message.event.music.snapcast.GroupStreamChangeEvent`
        * :class:`platypush.message.event.music.snapcast.StreamUpdateEvent`
        * :class:`platypush.message.event.music.snapcast.ServerUpdateEvent`
    """

    _DEFAULT_SNAPCAST_PORT = 1705
    _DEFAULT_POLL_SECONDS = 10   # Poll servers each 10 seconds
    _SOCKET_EOL = '\r\n'.encode()

    def __init__(self, hosts=None, ports=None,
                 poll_seconds=_DEFAULT_POLL_SECONDS, *args, **kwargs):
        """
        :param hosts: List of Snapcast server names or IPs to monitor (default:
            `['localhost']`
        :type hosts: list[str]

        :param ports: List of control ports for the configured Snapcast servers
            (default: `[1705]`)
        :type ports: list[int]

        :param poll_seconds: How often the backend will poll remote servers for
            status updated (default: 10 seconds)
        :type poll_seconds: float
        """

        super().__init__(*args, **kwargs)

        if hosts is None:
            hosts = ['localhost']
        if ports is None:
            ports = [self._DEFAULT_SNAPCAST_PORT]

        self.hosts = hosts[:]
        self.ports = ports[:]
        self.poll_seconds = poll_seconds
        self._socks = {}
        self._threads = {}
        self._statuses = {}

        if len(hosts) > len(ports):
            for _ in range(len(ports), len(hosts)):
                self.ports.append(self._DEFAULT_SNAPCAST_PORT)

    def _connect(self, host, port):
        if self._socks.get(host):
            return self._socks[host]

        self.logger.debug('Connecting to {host}:{port}'.format(host=host, port=port))
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))
        self._socks[host] = sock
        self.logger.info('Connected to {host}:{port}'.format(host=host, port=port))
        return sock

    def _disconnect(self, host, port):
        sock = self._socks.get(host)
        if not sock:
            self.logger.debug('Not connected to {}:{}'.format(host, port))
            return

        try:
            sock.close()
        except Exception as e:
            self.logger.warning(('Exception while disconnecting from {host}:{port}: {error}'.
                                 format(host=host, port=port, error=str(e))))
        finally:
            self._socks[host] = None

    @classmethod
    def _recv(cls, sock):
        buf = b''
        while buf[-2:] != cls._SOCKET_EOL:
            buf += sock.recv(1)
        return json.loads(buf.decode().strip())

    @classmethod
    def _parse_msg(cls, host, msg):
        evt = None

        if msg.get('method') == 'Client.OnVolumeChanged':
            client_id = msg.get('params', {}).get('id')
            volume = msg.get('params', {}).get('volume', {}).get('percent')
            muted = msg.get('params', {}).get('volume', {}).get('muted')
            evt = ClientVolumeChangeEvent(host=host, client=client_id,
                                          volume=volume, muted=muted)
        elif msg.get('method') == 'Group.OnMute':
            group_id = msg.get('params', {}).get('id')
            muted = msg.get('params', {}).get('mute')
            evt = GroupMuteChangeEvent(host=host, group=group_id, muted=muted)
        elif msg.get('method') == 'Client.OnConnect':
            client = msg.get('params', {}).get('client')
            evt = ClientConnectedEvent(host=host, client=client)
        elif msg.get('method') == 'Client.OnDisconnect':
            client = msg.get('params', {}).get('client')
            evt = ClientDisconnectedEvent(host=host, client=client)
        elif msg.get('method') == 'Client.OnLatencyChanged':
            client = msg.get('params', {}).get('id')
            latency = msg.get('params', {}).get('latency')
            evt = ClientLatencyChangeEvent(host=host, client=client, latency=latency)
        elif msg.get('method') == 'Client.OnNameChanged':
            client = msg.get('params', {}).get('id')
            name = msg.get('params', {}).get('name')
            evt = ClientNameChangeEvent(host=host, client=client, name=name)
        elif msg.get('method') == 'Group.OnStreamChanged':
            group_id = msg.get('params', {}).get('id')
            stream_id = msg.get('params', {}).get('stream_id')
            evt = GroupStreamChangeEvent(host=host, group=group_id, stream=stream_id)
        elif msg.get('method') == 'Stream.OnUpdate':
            stream_id = msg.get('params', {}).get('stream_id')
            stream = msg.get('params', {}).get('stream')
            evt = StreamUpdateEvent(host=host, stream_id=stream_id, stream=stream)
        elif msg.get('method') == 'Server.OnUpdate':
            server = msg.get('params', {}).get('server')
            evt = ServerUpdateEvent(host=host, server=server)

        return evt

    def _client(self, host, port, thread_name):
        def _thread():
            set_thread_name(thread_name)

            while not self.should_stop():
                try:
                    sock = self._connect(host, port)
                    msgs = self._recv(sock)

                    if not isinstance(msgs, list):
                        msgs = [msgs]

                    for msg in msgs:
                        self.logger.debug('Received message on {host}:{port}: {msg}'.
                                          format(host=host, port=port, msg=msg))

                        evt = self._parse_msg(host=host, msg=msg)
                        if evt:
                            self.bus.post(evt)
                except Exception as e:
                    self.logger.warning('Exception while getting the status ' + 'of the Snapcast server {}:{}: {}'.
                                        format(host, port, str(e)))

                    self._disconnect(host, port)
                finally:
                    time.sleep(self.poll_seconds)

        return _thread

    def run(self):
        super().run()

        self.logger.info('Initialized Snapcast backend - hosts: {} ports: {}'.
                         format(self.hosts, self.ports))

        while not self.should_stop():
            for i, host in enumerate(self.hosts):
                port = self.ports[i]
                thread_name = 'Snapcast-{host}-{port}'.format(host=host, port=port)

                self._threads[host] = threading.Thread(
                    target=self._client(host, port, thread_name),
                    name=thread_name
                )

                self._threads[host].start()

            for host in self.hosts:
                self._threads[host].join()


# vim:sw=4:ts=4:et:



import os
import struct
from typing import Optional, List

from platypush.message.response.stt import SpeechDetectedResponse
from platypush.plugins import action
from platypush.plugins.stt import SttPlugin


class SttPicovoiceHotwordPlugin(SttPlugin):
    """
    This plugin performs hotword detection using `PicoVoice <https://github.com/Picovoice>`_.

    Requires:

        * **pvporcupine** (``pip install pvporcupine``) for hotword detection.

    """

    def __init__(self,
                 library_path: Optional[str] = None,
                 model_file_path: Optional[str] = None,
                 keyword_file_paths: Optional[List[str]] = None,
                 sensitivity: float = 0.5,
                 sensitivities: Optional[List[float]] = None,
                 *args, **kwargs):
        from pvporcupine import Porcupine
        from pvporcupine.resources.util.python.util import LIBRARY_PATH, MODEL_FILE_PATH, KEYWORD_FILE_PATHS
        super().__init__(*args, **kwargs)

        self.hotwords = list(self.hotwords)
        self._hotword_engine: Optional[Porcupine] = None
        self._library_path = os.path.abspath(os.path.expanduser(library_path or LIBRARY_PATH))
        self._model_file_path = os.path.abspath(os.path.expanduser(model_file_path or MODEL_FILE_PATH))

        if not keyword_file_paths:
            hotwords = KEYWORD_FILE_PATHS
            assert all(hotword in hotwords for hotword in self.hotwords), \
                'Not all the hotwords could be found. Available hotwords: {}'.format(list(hotwords.keys()))

            self._keyword_file_paths = [os.path.abspath(os.path.expanduser(hotwords[hotword]))
                                        for hotword in self.hotwords]
        else:
            self._keyword_file_paths = [
                os.path.abspath(os.path.expanduser(p))
                for p in keyword_file_paths
            ]

        self._sensitivities = []
        if sensitivities:
            assert len(self._keyword_file_paths) == len(sensitivities), \
                'Please specify as many sensitivities as the number of configured hotwords'

            self._sensitivities = sensitivities
        else:
            self._sensitivities = [sensitivity] * len(self._keyword_file_paths)

    def convert_frames(self, frames: bytes) -> tuple:
        assert self._hotword_engine, 'The hotword engine is not running'
        return struct.unpack_from("h" * self._hotword_engine.frame_length, frames)

    def on_detection_ended(self) -> None:
        if self._hotword_engine:
            self._hotword_engine.delete()
        self._hotword_engine = None

    def detect_speech(self, frames: tuple) -> str:
        index = self._hotword_engine.process(frames)
        if index < 0:
            return ''

        if index is True:
            index = 0
        return self.hotwords[index]

    @action
    def detect(self, audio_file: str) -> SpeechDetectedResponse:
        """
        Perform speech-to-text analysis on an audio file.

        :param audio_file: Path to the audio file.
        """
        pass

    def recording_thread(self, input_device: Optional[str] = None, *args, **kwargs) -> None:
        assert self._hotword_engine, 'The hotword engine has not yet been initialized'
        super().recording_thread(block_size=self._hotword_engine.frame_length, input_device=input_device)

    @action
    def start_detection(self, *args, **kwargs) -> None:
        from pvporcupine import Porcupine
        self._hotword_engine = Porcupine(
            library_path=self._library_path,
            model_file_path=self._model_file_path,
            keyword_file_paths=self._keyword_file_paths,
            sensitivities=self._sensitivities)

        self.rate = self._hotword_engine.sample_rate
        super().start_detection(*args, **kwargs)


# vim:sw=4:ts=4:et:



from typing import Dict, Optional

from platypush.backend.sensor import SensorBackend
from platypush.message.event.bluetooth import BluetoothDeviceFoundEvent, BluetoothDeviceLostEvent


class BluetoothScannerBackend(SensorBackend):
    """
    This backend periodically scans for available bluetooth devices and returns events when a devices enter or exits
    the range.

    Triggers:

        * :class:`platypush.message.event.bluetooth.BluetoothDeviceFoundEvent` when a new bluetooth device is found.
        * :class:`platypush.message.event.bluetooth.BluetoothDeviceLostEvent` when a bluetooth device is lost.

    Requires:

        * The :class:`platypush.plugins.bluetooth.BluetoothPlugin` plugin working.

    """

    def __init__(self, device_id: Optional[int] = None, scan_duration: int = 10, **kwargs):
        """
        :param device_id: Bluetooth adapter ID to use (default configured on the ``bluetooth`` plugin if None).
        :param scan_duration:  How long the scan should run (default: 10 seconds).
        """
        super().__init__(plugin='bluetooth', plugin_args={
            'device_id': device_id,
            'duration': scan_duration,
        }, **kwargs)

        self._last_seen_devices = {}

    def process_data(self, data: Dict[str, dict], new_data: Dict[str, dict]):
        for addr, dev in data.items():
            if addr not in self._last_seen_devices:
                self.bus.post(BluetoothDeviceFoundEvent(address=dev.pop('addr'), **dev))
            self._last_seen_devices[addr] = {'addr': addr, **dev}

        for addr, dev in self._last_seen_devices.copy().items():
            if addr not in data:
                self.bus.post(BluetoothDeviceLostEvent(address=dev.pop('addr'), **dev))
                del self._last_seen_devices[addr]


# vim:sw=4:ts=4:et:



import enum
import functools
import os
import queue
import re
from typing import Optional, List, Dict, Union

import requests
import subprocess
import tempfile
import threading

from platypush.config import Config
from platypush.context import get_plugin, get_backend
from platypush.plugins import Plugin, action


class PlayerState(enum.Enum):
    STOP = 'stop'
    PLAY = 'play'
    PAUSE = 'pause'
    IDLE = 'idle'


class MediaPlugin(Plugin):
    """
    Generic plugin to interact with a media player.

    Requires:

        * A media player installed (supported so far: mplayer, vlc, mpv, omxplayer, chromecast)
        * **python-libtorrent** (``pip install python-libtorrent``), optional, for torrent support over native library
        * **youtube-dl** installed on your system (see your distro instructions), optional for YouTube support
        * **requests** (``pip install requests``), optional, for local files over HTTP streaming supporting
        * **ffmpeg**,optional, to get media files metadata

    To start the local media stream service over HTTP you will also need the
    :class:`platypush.backend.http.HttpBackend` backend enabled.
    """

    # A media plugin can either be local or remote (e.g. control media on
    # another device)
    _is_local = True
    _youtube_fifo = os.path.join(tempfile.gettempdir(), 'youtube_video.sock')

    _NOT_IMPLEMENTED_ERR = NotImplementedError(
        'This method must be implemented in a derived class')

    # Supported audio extensions
    audio_extensions = {
        '3gp', 'aa', 'aac', 'aax', 'act', 'aiff', 'amr', 'ape', 'au',
        'awb', 'dct', 'dss', 'dvf', 'flac', 'gsm', 'iklax', 'ivs',
        'm4a', 'm4b', 'm4p', 'mmf', 'mp3', 'mpc', 'msv', 'nmf', 'nsf',
        'ogg,', 'opus', 'ra,', 'raw', 'sln', 'tta', 'vox', 'wav',
        'wma', 'wv', 'webm', '8svx',
    }

    # Supported video extensions
    video_extensions = {
        'webm', 'mkv', 'flv', 'flv', 'vob', 'ogv', 'ogg', 'drc', 'gif',
        'gifv', 'mng', 'avi', 'mts', 'm2ts', 'mov', 'qt', 'wmv', 'yuv',
        'rm', 'rmvb', 'asf', 'amv', 'mp4', 'm4p', 'm4v', 'mpg', 'mp2',
        'mpeg', 'mpe', 'mpv', 'mpg', 'mpeg', 'm2v', 'm4v', 'svi',
        '3gp', '3g2', 'mxf', 'roq', 'nsv', 'flv', 'f4v', 'f4p', 'f4a',
        'f4b',
    }

    _supported_media_plugins = {'media.mplayer', 'media.omxplayer', 'media.mpv',
                                'media.vlc', 'media.chromecast'}

    _supported_media_types = ['file', 'torrent', 'youtube']
    _default_search_timeout = 60  # 60 seconds

    def __init__(self,
                 media_dirs: Optional[List[str]] = None,
                 download_dir: Optional[str] = None,
                 env: Optional[Dict[str, str]] = None,
                 volume: Optional[Union[float, int]] = None,
                 *args, **kwargs):
        """
        :param media_dirs: Directories that will be scanned for media files when
            a search is performed (default: none)

        :param download_dir: Directory where external resources/torrents will be
            downloaded (default: ~/Downloads)

        :param env: Environment variables key-values to pass to the
            player executable (e.g. DISPLAY, XDG_VTNR, PULSE_SINK etc.)

        :param volume: Default volume for the player (default: None, maximum volume).
        """

        super().__init__(**kwargs)

        if media_dirs is None:
            media_dirs = []
        player = None
        player_config = {}

        if self.__class__.__name__ == 'MediaPlugin':
            # Abstract class, initialize with the default configured player
            for plugin in Config.get_plugins().keys():
                if plugin in self._supported_media_plugins:
                    player = plugin
                    if get_plugin(player).is_local():
                        # Local players have priority as default if configured
                        break
        else:
            player = self  # Derived concrete class

        if not player:
            raise AttributeError('No media plugin configured')

        media_dirs = media_dirs or player_config.get('media_dirs', [])

        if self.__class__.__name__ == 'MediaPlugin':
            # Populate this plugin with the actions of the configured player
            plugin = get_plugin(player)
            for act in plugin.registered_actions:
                setattr(self, act, getattr(plugin, act))
                self.registered_actions.add(act)

        self._env = env or {}
        self.media_dirs = set(
            filter(
                lambda _: os.path.isdir(_),
                map(
                    lambda _: os.path.abspath(os.path.expanduser(_)),
                    media_dirs
                )
            )
        )

        self.download_dir = os.path.abspath(os.path.expanduser(
            download_dir or player_config.get('download_dir') or
            os.path.join((os.path.expanduser('~') or self._env.get('HOME') or '/'), 'Downloads')))

        if not os.path.isdir(self.download_dir):
            os.makedirs(self.download_dir, exist_ok=True)

        self.media_dirs.add(self.download_dir)
        self.volume = volume
        self._videos_queue = []
        self._youtube_proc = None

    @staticmethod
    def _torrent_event_handler(evt_queue):
        def handler(event):
            # More than 5% of the torrent has been downloaded
            if event.args.get('progress', 0) > 5 and event.args.get('files'):
                evt_queue.put(event.args['files'])
        return handler

    def _get_resource(self, resource):
        """
        :param resource: Resource to play/parse. Supported types:

            * Local files (format: ``file://<path>/<file>``)
            * Remote videos (format: ``https://<url>/<resource>``)
            * YouTube videos (format: ``https://www.youtube.com/watch?v=<id>``)
            * Torrents (format: Magnet links, Torrent URLs or local Torrent files)
        """

        if resource.startswith('youtube:') \
                or resource.startswith('https://youtu.be/') \
                or resource.startswith('https://www.youtube.com/watch?v='):
            m = re.match('youtube:video:(.*)', resource)
            if not m:
                m = re.match('https://youtu.be/(.*)', resource)
            if m:
                resource = 'https://www.youtube.com/watch?v={}'.format(m.group(1))

            if self.__class__.__name__ == 'MediaChromecastPlugin':
                # The Chromecast has already its native way to handle YouTube
                return resource

            self.stream_youtube_to_fifo(resource)
            resource = 'file://' + self._youtube_fifo
        elif resource.startswith('magnet:?'):
            self.logger.info('Downloading torrent {} to {}'.format(
                resource, self.download_dir))
            torrents = get_plugin('torrent')

            evt_queue = queue.Queue()
            torrents.download(resource, download_dir=self.download_dir, _async=True, is_media=True,
                              event_hndl=self._torrent_event_handler(evt_queue))

            resources = [f for f in evt_queue.get()]

            if resources:
                self._videos_queue = sorted(resources)
                resource = self._videos_queue.pop(0)
            else:
                raise RuntimeError('No media file found in torrent {}'.format(resource))

        return resource

    @staticmethod
    def _stop_torrent():
        torrents = get_plugin('torrent')
        torrents.quit()

    @action
    def play(self, resource, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def pause(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def stop(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def quit(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def voldown(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def volup(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def back(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def forward(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def next(self):
        """ Play the next item in the queue """
        self.stop()

        if self._videos_queue:
            video = self._videos_queue.pop(0)
            return self.play(video)

    @action
    def toggle_subtitles(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def set_subtitles(self, filename, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def remove_subtitles(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def is_playing(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def load(self, resource, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def mute(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def seek(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def set_position(self, *args, **kwargs):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def set_volume(self, volume):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def status(self):
        raise self._NOT_IMPLEMENTED_ERR

    @action
    def search(self, query, types=None, queue_results=False, autoplay=False,
               search_timeout=_default_search_timeout):
        """
        Perform a video search.

        :param query: Query string, video name or partial name
        :type query: str

        :param types: Video types to search (default: ``["youtube", "file", "torrent"]``)
        :type types: list

        :param queue_results: Append the results to the current playing queue (default: False)
        :type queue_results: bool

        :param autoplay: Play the first result of the search (default: False)
        :type autoplay: bool

        :param search_timeout: Search timeout (default: 60 seconds)
        :type search_timeout: float
        """

        results = {}
        results_queues = {}
        worker_threads = {}

        if types is None:
            types = self._supported_media_types

        for media_type in types:
            results[media_type] = []
            results_queues[media_type] = queue.Queue()
            search_hndl = self._get_search_handler_by_type(media_type)
            worker_threads[media_type] = threading.Thread(
                target=self._search_worker(query=query, search_hndl=search_hndl,
                                           results_queue=results_queues[media_type]))
            worker_threads[media_type].start()

        for media_type in types:
            try:
                results[media_type].extend(
                    results_queues[media_type].get(timeout=search_timeout))
            except queue.Empty:
                self.logger.warning('Search for "{}" media type {} timed out'.
                                    format(query, media_type))

        flattened_results = []

        for media_type in self._supported_media_types:
            if media_type in results:
                for result in results[media_type]:
                    result['type'] = media_type
                flattened_results += results[media_type]

        results = flattened_results

        if results:
            if queue_results:
                self._videos_queue = [_['url'] for _ in results]
                if autoplay:
                    self.play(self._videos_queue.pop(0))
            elif autoplay:
                self.play(results[0]['url'])

        return results

    @staticmethod
    def _search_worker(query, search_hndl, results_queue):
        def thread():
            results_queue.put(search_hndl.search(query))
        return thread

    def _get_search_handler_by_type(self, search_type):
        if search_type == 'file':
            from .search import LocalMediaSearcher
            return LocalMediaSearcher(self.media_dirs)
        if search_type == 'torrent':
            from .search import TorrentMediaSearcher
            return TorrentMediaSearcher()
        if search_type == 'youtube':
            from .search import YoutubeMediaSearcher
            return YoutubeMediaSearcher()

        self.logger.warning('Unsupported search type: {}'.format(search_type))

    @classmethod
    def _is_video_file(cls, filename):
        return filename.lower().split('.')[-1] in cls.video_extensions

    @classmethod
    def _is_audio_file(cls, filename):
        return filename.lower().split('.')[-1] in cls.audio_extensions

    @action
    def start_streaming(self, media, subtitles=None, download=False):
        """
        Starts streaming local media over the specified HTTP port.
        The stream will be available to HTTP clients on
        `http://{this-ip}:{http_backend_port}/media/<media_id>`

        :param media: Media to stream
        :type media: str

        :param subtitles: Path or URL to the subtitles track to be used
        :type subtitles: str

        :param download: Set to True if you prefer to download the file from
            the streaming link instead of streaming it
        :type download: bool

        :return: dict containing the streaming URL.Example:

        .. code-block:: json

            {
                "id": "0123456abcdef.mp4",
                "source": "file:///mnt/media/movies/movie.mp4",
                "mime_type": "video/mp4",
                "url": "http://192.168.1.2:8008/media/0123456abcdef.mp4"
            }

        """
        import requests

        http = get_backend('http')
        if not http:
            self.logger.warning('Unable to stream {}: HTTP backend unavailable'.
                                format(media))
            return

        self.logger.info('Starting streaming {}'.format(media))
        response = requests.put('{url}/media{download}'.format(
            url=http.local_base_url, download='?download' if download else ''),
            json={'source': media, 'subtitles': subtitles})

        if not response.ok:
            self.logger.warning('Unable to start streaming: {}'.
                                format(response.text or response.reason))
            return None, (response.text or response.reason)

        return response.json()

    @action
    def stop_streaming(self, media_id):
        import requests

        http = get_backend('http')
        if not http:
            self.logger.warning('Cannot unregister {}: HTTP backend unavailable'.
                                format(media_id))
            return

        response = requests.delete('{url}/media/{id}'.
                                   format(url=http.local_base_url, id=media_id))

        if not response.ok:
            self.logger.warning('Unable to unregister media_id {}: {}'.format(
                media_id, response.reason))
            return

        return response.json()

    @staticmethod
    def _youtube_search_api(query):
        return [
            {
                'url': 'https://www.youtube.com/watch?v=' + item['id']['videoId'],
                'title': item.get('snippet', {}).get('title', '<No Title>'),
            }
            for item in get_plugin('google.youtube').search(query=query).output
            if item.get('id', {}).get('kind') == 'youtube#video'
        ]

    @staticmethod
    def _youtube_search_html_parse(query):
        from .search import YoutubeMediaSearcher
        # noinspection PyProtectedMember
        return YoutubeMediaSearcher()._youtube_search_html_parse(query)

    def stream_youtube_to_fifo(self, url):
        if self._youtube_proc:
            self.logger.info('Terminating existing YouTube process')
            self._youtube_proc.terminate()
            self._youtube_proc = None

        if os.path.exists(self._youtube_fifo):
            os.unlink(self._youtube_fifo)

        os.mkfifo(self._youtube_fifo, 0o644)

        def _youtube_dl_thread():
            self._youtube_proc = subprocess.Popen(['youtube-dl', '-f', 'best', '-o', self._youtube_fifo, url])
            self._youtube_proc.wait()
            self._youtube_proc = None

        threading.Thread(target=_youtube_dl_thread).start()

    @action
    def get_youtube_url(self, url):
        m = re.match('youtube:video:(.*)', url)
        if m:
            url = 'https://www.youtube.com/watch?v={}'.format(m.group(1))

        proc = subprocess.Popen(['youtube-dl', '-f', 'best', '-g', url], stdout=subprocess.PIPE)
        raw_url = proc.stdout.read().decode("utf-8", "strict")[:-1]
        return raw_url if raw_url else url

    @action
    def get_youtube_info(self, url):
        m = re.match('youtube:video:(.*)', url)
        if m:
            url = 'https://www.youtube.com/watch?v={}'.format(m.group(1))

        proc = subprocess.Popen(['youtube-dl', '-j', url], stdout=subprocess.PIPE)
        return proc.stdout.read().decode("utf-8", "strict")[:-1]

    @action
    def get_media_file_duration(self, filename):
        """
        Get the duration of a media file in seconds. Requires ffmpeg
        """

        if filename.startswith('file://'):
            filename = filename[7:]

        result = subprocess.Popen(["ffprobe", filename], stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

        return functools.reduce(
            lambda t, t_i: t + t_i,
            [float(t) * pow(60, i) for (i, t) in enumerate(re.search(
                '^Duration:\s*([^,]+)', [x.decode()
                                         for x in result.stdout.readlines()
                                         if "Duration" in x.decode()].pop().strip()
            ).group(1).split(':')[::-1])]
        )

    @action
    def download(self, url, filename=None, directory=None):
        """
        Download a media URL

        :param url: Media URL
        :param filename: Media filename (default: URL filename)
        :param directory: Destination directory (default: download_dir)
        :return: The absolute path to the downloaded file
        """

        if not filename:
            filename = url.split('/')[-1]
        if not directory:
            directory = self.download_dir

        path = os.path.join(directory, filename)
        content = requests.get(url).content

        with open(path, 'wb') as f:
            f.write(content)

        return path

    def is_local(self):
        return self._is_local

    @staticmethod
    def get_subtitles_file(subtitles):
        if not subtitles:
            return

        if subtitles.startswith('file://'):
            subtitles = subtitles[len('file://'):]
        if os.path.isfile(subtitles):
            return os.path.abspath(subtitles)
        else:
            content = requests.get(subtitles).content
            f = tempfile.NamedTemporaryFile(prefix='media_subs_',
                                            suffix='.srt', delete=False)

            with f:
                f.write(content)
            return f.name


# vim:sw=4:ts=4:et:



from platypush.plugins import Plugin, action


class SwitchPlugin(Plugin):
    """
    Abstract class for interacting with switch devices
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @action
    def on(self, device, *args, **kwargs):
        """ Turn the device on """
        raise NotImplementedError()

    @action
    def off(self, device, *args, **kwargs):
        """ Turn the device off """
        raise NotImplementedError()

    @action
    def toggle(self, device, *args, **kwargs):
        """ Toggle the device status (on/off) """
        raise NotImplementedError()

    @action
    def status(self, device=None, *args, **kwargs):
        """ Get the status of a specified device or of all the configured devices (default)"""
        devices = self.devices
        if device:
            devices = [d for d in self.devices if d.get('id') == device or d.get('name') == device]
            if devices:
                return self.devices.pop(0)
            else:
                return None

        return devices

    @property
    def devices(self):
        raise NotImplementedError()


# vim:sw=4:ts=4:et:



from typing import Optional, Dict

from platypush.plugins import Plugin, action


class GraphitePlugin(Plugin):
    """
    Plugin for sending data to a Graphite instance.
    """

    def __init__(self, host: str = 'localhost', port: int = 2003, timeout: int = 5, **kwargs):
        """
        :param host: Default Graphite host (default: 'localhost').
        :param port: Default Graphite port (default: 2003).
        :param timeout: Communication timeout in seconds (default: 5).
        """
        super().__init__(**kwargs)
        self.host = host
        self.port = port
        self.timeout = timeout

    @action
    def send(self,
             metric: str,
             value,
             host: Optional[str] = None,
             port: Optional[int] = None,
             timeout: Optional[int] = None,
             tags: Optional[Dict[str, str]] = None,
             prefix: str = '',
             protocol: str = 'tcp'):
        """
        Send data to a Graphite instance.

        :param metric: Metric name.
        :param value: Value to be sent.
        :param host: Graphite host (default: default configured ``host``).
        :param port: Graphite port (default: default configured ``port``).
        :param tags: Map of tags for the metric.
        :param prefix: Metric prefix name (default: empty string).
        :param protocol: Communication protocol - possible values: 'tcp', 'udp' (default: 'tcp').
        """
        import graphyte

        host = host or self.host
        port = port or self.port
        tags = tags or {}
        protocol = protocol.lower()

        graphyte.init(host, port=port, prefix=prefix, protocol=protocol, timeout=timeout or self.timeout)
        graphyte.send(metric, value, tags=tags)


# vim:sw=4:ts=4:et:



import time

import platypush.plugins.gpio.zeroborg.lib as ZeroBorg

no_msg_timeout = 0.37
last_msg = None
last_msg_timestamp = None
auto_mode = False

ZB = ZeroBorg.ZeroBorg()
ZB.Init()

while True:
    ZB.GetIrMessage()
    if ZB.HasNewIrMessage():
        message = ZB.GetIrMessage()
        print('Received message: {}'.format(message))

        last_msg = message
        last_msg_timestamp = time.time()

# vim:sw=4:ts=4:et:




"""Based on Google pushtotalk.py sample."""

import concurrent.futures
import json
import logging

import grpc

from google.assistant.embedded.v1alpha2 import (
    embedded_assistant_pb2,
    embedded_assistant_pb2_grpc
)
from tenacity import retry, stop_after_attempt, retry_if_exception

try:
    from googlesamples.assistant.grpc import (
        assistant_helpers,
        audio_helpers,
        browser_helpers,
        device_helpers
    )
except (SystemError, ImportError):
    import assistant_helpers
    import audio_helpers
    import browser_helpers
    import device_helpers


ASSISTANT_API_ENDPOINT = 'embeddedassistant.googleapis.com'
END_OF_UTTERANCE = embedded_assistant_pb2.AssistResponse.END_OF_UTTERANCE
DIALOG_FOLLOW_ON = embedded_assistant_pb2.DialogStateOut.DIALOG_FOLLOW_ON
CLOSE_MICROPHONE = embedded_assistant_pb2.DialogStateOut.CLOSE_MICROPHONE
PLAYING = embedded_assistant_pb2.ScreenOutConfig.PLAYING
DEFAULT_GRPC_DEADLINE = 60 * 3 + 5


class SampleAssistant(object):
    """Sample Assistant that supports conversations and device actions.

    Args:
      device_model_id: identifier of the device model.
      device_id: identifier of the registered device instance.
      conversation_stream(ConversationStream): audio stream
        for recording query and playing back assistant answer.
      channel: authorized gRPC channel for connection to the
        Google Assistant API.
      deadline_sec: gRPC deadline in seconds for Google Assistant API call.
      device_handler: callback for device actions.
    """

    def __init__(self, language_code, device_model_id, device_id,
                 conversation_stream, display,
                 channel, deadline_sec, device_handler, play_response=True,
                 on_conversation_start=None, on_conversation_end=None,
                 on_speech_recognized=None, on_volume_changed=None,
                 on_response=None):
        self.language_code = language_code
        self.device_model_id = device_model_id
        self.device_id = device_id
        self.conversation_stream = conversation_stream
        self.display = display
        self.play_response = play_response

        # Opaque blob provided in AssistResponse that,
        # when provided in a follow-up AssistRequest,
        # gives the Assistant a context marker within the current state
        # of the multi-Assist()-RPC "conversation".
        # This value, along with MicrophoneMode, supports a more natural
        # "conversation" with the Assistant.
        self.conversation_state = None
        # Force reset of first conversation.
        self.is_new_conversation = True

        # Create Google Assistant API gRPC client.
        self.assistant = embedded_assistant_pb2_grpc.EmbeddedAssistantStub(channel)
        self.deadline = deadline_sec

        self.device_handler = device_handler
        self.detected_speech = None

        self.on_conversation_start = on_conversation_start
        self.on_conversation_end = on_conversation_end
        self.on_speech_recognized = on_speech_recognized
        self.on_volume_changed = on_volume_changed
        self.on_response = on_response

    def __enter__(self):
        return self

    def __exit__(self, etype, e, traceback):
        if e:
            return False
        self.conversation_stream.close()

    @staticmethod
    def is_grpc_error_unavailable(e):
        is_grpc_error = isinstance(e, grpc.RpcError)
        if is_grpc_error and (e.code() == grpc.StatusCode.UNAVAILABLE):
            logging.error('grpc unavailable error: %s', e)
            return True
        return False

    @retry(reraise=True, stop=stop_after_attempt(3),
           retry=retry_if_exception(is_grpc_error_unavailable))
    def assist(self):
        """Send a voice request to the Assistant and playback the response.

        Returns: True if conversation should continue.
        """
        continue_conversation = False
        device_actions_futures = []

        self.conversation_stream.start_recording()
        if self.on_conversation_start:
            self.on_conversation_start()

        logging.info('Recording audio request.')

        def iter_log_assist_requests():
            for c in self.gen_assist_requests():
                assistant_helpers.log_assist_request_without_audio(c)
                yield c
            logging.debug('Reached end of AssistRequest iteration.')

        # This generator yields AssistResponse proto messages
        # received from the gRPC Google Assistant API.
        for resp in self.assistant.Assist(iter_log_assist_requests(), self.deadline):
            assistant_helpers.log_assist_response_without_audio(resp)
            if resp.event_type == END_OF_UTTERANCE:
                logging.info('End of audio request detected.')
                logging.info('Stopping recording.')
                self.conversation_stream.stop_recording()

                if self.detected_speech and self.on_speech_recognized:
                    self.on_speech_recognized(self.detected_speech)

            if resp.speech_results:
                self.detected_speech = ' '.join(
                    r.transcript.strip() for r in resp.speech_results
                    if len(r.transcript.strip())).strip()

                logging.info('Transcript of user request: "%s".', self.detected_speech)

            if len(resp.audio_out.audio_data) > 0:
                if not self.conversation_stream.playing:
                    self.conversation_stream.stop_recording()

                    if self.play_response:
                        self.conversation_stream.start_playback()
                        logging.info('Playing assistant response.')

                if self.play_response and self.conversation_stream.playing:
                    self.conversation_stream.write(resp.audio_out.audio_data)
                elif self.conversation_stream.playing:
                    self.conversation_stream.stop_playback()

            if resp.dialog_state_out.conversation_state:
                conversation_state = resp.dialog_state_out.conversation_state
                logging.debug('Updating conversation state.')
                self.conversation_state = conversation_state

            if resp.dialog_state_out.volume_percentage != 0:
                volume_percentage = resp.dialog_state_out.volume_percentage
                logging.info('Setting volume to %s%%', volume_percentage)
                self.conversation_stream.volume_percentage = volume_percentage

                if self.on_volume_changed:
                    self.on_volume_changed(volume_percentage)

            if resp.dialog_state_out.microphone_mode == DIALOG_FOLLOW_ON:
                continue_conversation = True
                logging.info('Expecting follow-on query from user.')
            elif resp.dialog_state_out.microphone_mode == CLOSE_MICROPHONE:
                continue_conversation = False

            if resp.device_action.device_request_json:
                device_request = json.loads(
                    resp.device_action.device_request_json
                )
                fs = self.device_handler(device_request)
                if fs:
                    device_actions_futures.extend(fs)

            if self.display and resp.screen_out.data:
                system_browser = browser_helpers.system_browser
                system_browser.display(resp.screen_out.data)

            if resp.dialog_state_out.supplemental_display_text and self.on_response:
                self.on_response(resp.dialog_state_out.supplemental_display_text)

        if len(device_actions_futures):
            logging.info('Waiting for device executions to complete.')
            concurrent.futures.wait(device_actions_futures)

        logging.info('Finished playing assistant response.')
        self.conversation_stream.stop_playback()

        if self.on_conversation_end:
            self.on_conversation_end(continue_conversation)

        return continue_conversation

    def gen_assist_requests(self):
        """Yields: AssistRequest messages to send to the API."""

        config = embedded_assistant_pb2.AssistConfig(
            audio_in_config=embedded_assistant_pb2.AudioInConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
            ),
            audio_out_config=embedded_assistant_pb2.AudioOutConfig(
                encoding='LINEAR16',
                sample_rate_hertz=self.conversation_stream.sample_rate,
                volume_percentage=self.conversation_stream.volume_percentage,
            ),
            dialog_state_in=embedded_assistant_pb2.DialogStateIn(
                language_code=self.language_code,
                conversation_state=self.conversation_state,
                is_new_conversation=self.is_new_conversation,
            ),
            device_config=embedded_assistant_pb2.DeviceConfig(
                device_id=self.device_id,
                device_model_id=self.device_model_id,
            )
        )
        if self.display:
            config.screen_out_config.screen_mode = PLAYING
        # Continue current conversation with later requests.
        self.is_new_conversation = False
        # The first AssistRequest must contain the AssistConfig
        # and no audio data.
        yield embedded_assistant_pb2.AssistRequest(config=config)
        for data in self.conversation_stream:
            # Subsequent requests need audio data, but not config.
            yield embedded_assistant_pb2.AssistRequest(audio_in=data)

